from collections.abc import Mapping
from typing import Any, TypeVar
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="OIDCPrincipal")


@_attrs_define
class OIDCPrincipal:
    """
    Attributes:
        organization_uuid (UUID):
        integration_uuid (UUID):
        principal_id (str):
    """

    organization_uuid: UUID
    integration_uuid: UUID
    principal_id: str
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        organization_uuid = str(self.organization_uuid)

        integration_uuid = str(self.integration_uuid)

        principal_id = self.principal_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "organization_uuid": organization_uuid,
                "integration_uuid": integration_uuid,
                "principal_id": principal_id,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        organization_uuid = UUID(d.pop("organization_uuid"))

        integration_uuid = UUID(d.pop("integration_uuid"))

        principal_id = d.pop("principal_id")

        oidc_principal = cls(
            organization_uuid=organization_uuid,
            integration_uuid=integration_uuid,
            principal_id=principal_id,
        )

        oidc_principal.additional_properties = d
        return oidc_principal

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import argparse
from invariant_client.bindings.invariant_instance_client.models.document_type import DocumentType
from invariant_client.bindings.invariant_instance_client.models.resource_type import ResourceType
from invariant_client.base_command.base_command import BaseCommand

class LocationsCommand(BaseCommand):
    use_argument_debug = True
    use_argument_group_format = True
    use_argument_format_tsv = True
    
    def __init__(self):
        super().__init__()
        self.network_name: str | None = None
        self.file_path: str | None = None
    
    @classmethod
    def parse_args(cls, subparsers: 'argparse._SubParsersAction[argparse.ArgumentParser]') -> None:
        command_locations = subparsers.add_parser(
            'locations',
            description='List or manage locations.',
            help='List or manage locations.'
        )
        cls._add_common_parser_arguments(command_locations)
        locations_subparsers = command_locations.add_subparsers(
            dest='locations_subcommand',
            required=True,
            help='Available subcommands for locations'
        )
        import_parser = locations_subparsers.add_parser(
            'import',
            description='Import locations from a file.',
            help='Import locations from a file.'
        )
        import_parser.add_argument(
            'network_name',
            help='The name of the network to import locations into.'
        )
        import_parser.add_argument(
            'file_path',
            help='Path to the file containing locations to import.'
        )
        cls._add_common_parser_arguments(import_parser)


    def set_config(self, args: 'argparse.Namespace', env: dict[str, str]) -> None:
        super().set_config(args, env)
        self.network_name = getattr(args, 'network_name', None)
        self.file_path = getattr(args, 'file_path', None)

    def execute(self) -> None:
        super().execute()
        self.sdk.import_editable_document(f"Network-Locations-{self.network_name}", DocumentType.INVARIANT_YAML_SPEC, ResourceType.LOCATIONS, self.file_path)


from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.rule_outcome import RuleOutcome
from ..models.rule_type import RuleType
from ..types import UNSET, Unset

T = TypeVar("T", bound="PolicyFileResultRule")


@_attrs_define
class PolicyFileResultRule:
    """
    Attributes:
        status (RuleOutcome): Rule evaluation outcome.
        rule_type (Union[None, RuleType]):
        comment (Union[None, Unset, str]):
        errors (Union[None, Unset, list[str]]):
    """

    status: RuleOutcome
    rule_type: Union[None, RuleType]
    comment: Union[None, Unset, str] = UNSET
    errors: Union[None, Unset, list[str]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        status = self.status.value

        rule_type: Union[None, str]
        if isinstance(self.rule_type, RuleType):
            rule_type = self.rule_type.value
        else:
            rule_type = self.rule_type

        comment: Union[None, Unset, str]
        if isinstance(self.comment, Unset):
            comment = UNSET
        else:
            comment = self.comment

        errors: Union[None, Unset, list[str]]
        if isinstance(self.errors, Unset):
            errors = UNSET
        elif isinstance(self.errors, list):
            errors = self.errors

        else:
            errors = self.errors

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "status": status,
                "rule_type": rule_type,
            }
        )
        if comment is not UNSET:
            field_dict["comment"] = comment
        if errors is not UNSET:
            field_dict["errors"] = errors

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        status = RuleOutcome(d.pop("status"))

        def _parse_rule_type(data: object) -> Union[None, RuleType]:
            if data is None:
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                rule_type_type_0 = RuleType(data)

                return rule_type_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, RuleType], data)

        rule_type = _parse_rule_type(d.pop("rule_type"))

        def _parse_comment(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        comment = _parse_comment(d.pop("comment", UNSET))

        def _parse_errors(data: object) -> Union[None, Unset, list[str]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                errors_type_0 = cast(list[str], data)

                return errors_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[str]], data)

        errors = _parse_errors(d.pop("errors", UNSET))

        policy_file_result_rule = cls(
            status=status,
            rule_type=rule_type,
            comment=comment,
            errors=errors,
        )

        policy_file_result_rule.additional_properties = d
        return policy_file_result_rule

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import logging
import json
import pathlib
from typing import Dict, Union

import boto3
import botocore
from botocore.exceptions import ClientError
from mypy_boto3_ec2 import EC2Client
from mypy_boto3_elbv2 import ElasticLoadBalancingv2Client
from mypy_boto3_rds import RDSClient

logger = logging.getLogger(__name__)

class AWSClientError(Exception):
    """Raised when AWS client is unable to be created"""


class AWSClient:
    _session = None
    resources = {
        "addresses": {"client": "ec2", "function": "describe_addresses"},
        "availability_zones": {
            "client": "ec2",
            "function": "describe_availability_zones",
        },
        "customer_gateways": {
            "client": "ec2",
            "function": "describe_customer_gateways",
        },
        "internet_gateways": {
            "client": "ec2",
            "function": "describe_internet_gateways",
        },
        "nat_gateways": {"client": "ec2", "function": "describe_nat_gateways"},
        "network_acls": {"client": "ec2", "function": "describe_network_acls"},
        "network_interfaces": {
            "client": "ec2",
            "function": "describe_network_interfaces",
        },
        "prefix_lists": {"client": "ec2", "function": "describe_prefix_lists"},
        "instances": {"client": "ec2", "function": "describe_instances"},
        "route_tables": {"client": "ec2", "function": "describe_route_tables"},
        "security_groups": {"client": "ec2", "function": "describe_security_groups"},
        "subnets": {"client": "ec2", "function": "describe_subnets"},
        "transit_gateway_attachments": {
            "client": "ec2",
            "function": "describe_transit_gateway_attachments",
        },
        "transit_gateway_route_tables": {
            "client": "ec2",
            "function": "describe_transit_gateway_route_tables",
        },
        "transit_gateway_vpc_attachments": {
            "client": "ec2",
            "function": "describe_transit_gateway_vpc_attachments",
        },
        "transit_gateways": {"client": "ec2", "function": "describe_transit_gateways"},
        "vpc_endpoints": {"client": "ec2", "function": "describe_vpc_endpoints"},
        "vpc_peering_connections": {
            "client": "ec2",
            "function": "describe_vpc_peering_connections",
        },
        "vpcs": {"client": "ec2", "function": "describe_vpcs"},
        "vpn_connections": {"client": "ec2", "function": "describe_vpn_connections"},
        "vpn_gateways": {"client": "ec2", "function": "describe_vpn_gateways"},
        "db_instances": {"client": "rds", "function": "describe_db_instances"},
        "load_balancers": {"client": "elbv2", "function": "describe_load_balancers"},
        "target_groups": {"client": "elbv2", "function": "describe_target_groups"},
    }

    def __init__(
        self,
        accounts=[],
        ignore_accounts=[],
        regions=set([]),
        profile=None,
        role=None,
        skip_resources=[],
    ):
        self.fatal = False
        self.profile = profile
        self.role = role
        if self.profile:
            try:
                self._session = boto3.session.Session(profile_name=self.profile)
            except botocore.exceptions.ProfileNotFound as e:
                raise ConnectionError(f"Unable to setup AWS session: {e}")
        else:
            self._session = boto3.session.Session()
        self.regions = set(regions if regions else self._get_regions())
        self.accounts = self._get_accounts(accounts, ignore_accounts)
        for resource in skip_resources:
            del self.resources[resource]

    def _get_regions(self):
        account_client = self._session.client("account")
        try:
            result = [
                i.get("RegionName")
                for i in account_client.list_regions(
                    RegionOptStatusContains=["ENABLED", "ENABLED_BY_DEFAULT"]
                )["Regions"]
            ]
        except ClientError as e:
            raise ConnectionError(f"Unable to get regions: {e}")
        return set(result)

    def _get_accounts(self, include_accounts=[], exclude_accounts=[]) -> Dict:
        """Get all accessible AWS accounts and optionally filter them.

        Args:
          include_accounts: Account IDs to include, if empty, include all.
          exclude_accounts: Account IDs to exclude.

        Returns:
          A list of accounts to return.
        """
        try:
            org_client = boto3.client("organizations")
            all_accounts = (
                org_client.get_paginator("list_accounts")
                .paginate()
                .build_full_result()["Accounts"]
            )
        except ClientError as e:
            raise AWSClientError(f"Unable to get organization accounts: {e}")

        filtered_accounts = []
        for account_dict in all_accounts:
            account_id = account_dict.get("Id")

            # Inclusion Logic:
            if include_accounts:
                if account_id not in include_accounts:
                    continue  # Skip if not in include list

            # Exclusion Logic:
            if exclude_accounts:
                if account_id in exclude_accounts:
                    continue  # Skip if in exclude list

            filtered_accounts.append(account_dict)
        return filtered_accounts

    def _assume_role(self, account_id, role) -> boto3.session.Session:
        """Assume an account role within AWS.

        Args:
          account_id: Account ID to attempt to assume role on.
          role: Role name to assume.

        Returns:
          Boto3 session

        Raises:
          AWSClientError: When a role is unable to be assumed.
        """
        try:
            sts_client = boto3.client("sts")
            assumed_role = sts_client.assume_role(
                RoleArn=f"arn:aws:iam::{account_id}:role/{role}",
                RoleSessionName="InvariantClient",
            )
            session = boto3.session.Session(
                aws_access_key_id=assumed_role["Credentials"]["AccessKeyId"],
                aws_secret_access_key=assumed_role["Credentials"]["SecretAccessKey"],
                aws_session_token=assumed_role["Credentials"]["SessionToken"],
            )
            return session
        except ClientError as e:
            raise AWSClientError(
                f"Error assuming role for account {account_id}, you can add this to `ignore_accounts` to suppress this error, skipping account: {e}"
            )

    def write_data(self, path, account, region, name, data):
        aws_dir = pathlib.Path(path) / "aws_configs"
        aws_dir.parent.mkdir(parents=True, exist_ok=True)
        config_with_ext = name + ".json"
        config_path = aws_dir / account / region / config_with_ext
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(config_path, "w") as f:
            json.dump(
                data,
                f,
                indent=1,
                default=str,
                sort_keys=True,
            )

    def _retrieve_configs(
        self,
        client: Union[EC2Client, RDSClient, ElasticLoadBalancingv2Client],
        function_name: str,
        account: str,
        options: dict[str, str] = {},
        path: str = "",
    ):
        """Retrieves config

        Args:
          client: A boto3 client specific to the functions to be called.
          functions: A list of strings of the functions to be called.
          account: The account ID to retrieve configs from.
          options: Options to pass into paginator.

        Raises:
          AWSClientError: Raised when client is unable to be created.
        """
        logger.info(
            f"Fetching {function_name} on account {account} in region {client.meta.region_name}."
        )
        try:
            if client.can_paginate(function_name):
                results = (
                    client.get_paginator(function_name)
                    .paginate(**options)
                    .build_full_result()
                )
            else:
                results = getattr(client, function_name, None)()
                del results["ResponseMetadata"]
            if len(results.keys()) != 1:
                self.fatal = True
                logger.warning(
                    f"Error processing {function_name} results, expected 1 key but got {len(results.keys())}: {results})"
                )
                return
            config_name = next(iter(results))
            self.write_data(
                path, account, client.meta.region_name, config_name, results
            )
        except ClientError as e:
            self.fatal = True
            logger.error(f"Fatal Error: {e}")
            return
        if function_name == "describe_transit_gateway_route_tables":
            tgw_rt = results["TransitGatewayRouteTables"]
            props = self._retrieve_transit_gateway_propagations(client, tgw_rt)
            self.write_data(
                path,
                account,
                client.meta.region_name,
                "TransitGatewayPropagations",
                props,
            )
            routes = self._retrieve_transit_gateway_static_routes(client, tgw_rt)
            self.write_data(
                path,
                account,
                client.meta.region_name,
                "TransitGatewayStaticRoutes",
                routes,
            )
        elif function_name == "describe_load_balancers":
            lbs = results["LoadBalancers"]
            listeners = self._retrieve_load_balancer_listeners(
                client, account, client.meta.region_name, lbs
            )
            self.write_data(
                path,
                account,
                client.meta.region_name,
                "LoadBalancerListeners",
                listeners,
            )
            attributes = self._retrieve_load_balancer_attributes(
                client, account, client.meta.region_name, lbs
            )
            self.write_data(
                path,
                account,
                client.meta.region_name,
                "LoadBalancerAttributes",
                attributes,
            )
        elif function_name == "describe_target_groups":
            target_groups = results["TargetGroups"]
            groups = self._retrieve_load_balancer_target_groups(
                client, account, client.meta.region_name, target_groups
            )
            self.write_data(
                path,
                account,
                client.meta.region_name,
                "LoadBalancerTargetHealth",
                groups,
            )

    def _retrieve_elastic_search_domains(
        self, client: ElasticLoadBalancingv2Client, account: str, path: str
    ):
        domain_names = client.list_domain_names()
        elastic_search_config = client.describe_elasticsearch_domains(
            DomainNames=[
                domainEntry["DomainName"] for domainEntry in domain_names["DomainNames"]
            ]
        )
        del elastic_search_config["ResponseMetadata"]
        self.write_data(
            path,
            account,
            client.meta.region_name,
            "ElasticsearchDomains",
            elastic_search_config,
        )

    def _retrieve_transit_gateway_propagations(self, client: EC2Client, routes):
        props = []
        logger.info(
            "Fetching get_transit_gateway_route_table_propogations for each transit gateway route table id."
        )
        for route in routes:
            try:
                result = client.get_transit_gateway_route_table_propagations(
                    TransitGatewayRouteTableId=route["TransitGatewayRouteTableId"]
                )
                result["TransitGatewayRouteTableId"] = route[
                    "TransitGatewayRouteTableId"
                ]
                del result["ResponseMetadata"]
            except ClientError as e:
                self.fatal = True
                logger.error(e)
                return
            props.append(result)
        return {"TransitGatewayPropagations": props}

    def _retrieve_transit_gateway_static_routes(self, client: EC2Client, routes):
        props = []
        logger.info(
            "Fetching search_transit_gateway_routes for each transit gateway route table id."
        )
        for route in routes:
            try:
                result = client.search_transit_gateway_routes(
                    TransitGatewayRouteTableId=route["TransitGatewayRouteTableId"],
                    Filters=[{"Name": "type", "Values": ["static"]}],
                )
                result["TransitGatewayRouteTableId"] = route[
                    "TransitGatewayRouteTableId"
                ]
                del result["ResponseMetadata"]
            except ClientError as e:
                self.fatal = True
                logger.error(e)
                return
            props.append(result)
        return {"TransitGatewayStaticRoutes": props}

    def _retrieve_load_balancer_target_groups(
        self, client: ElasticLoadBalancingv2Client, account_id, region, target_groups
    ):
        responses = []
        logger.info("Fetching describe_target_health for each ELB target group.")
        for group in target_groups:
            try:
                arn = group["TargetGroupArn"]
                resp = client.describe_target_health(TargetGroupArn=arn)
                resp["TargetGroupArn"] = group["TargetGroupArn"]
                del resp["ResponseMetadata"]
            except (ClientError, KeyError) as e:
                self.fatal = True
                logger.error(e)
                return
            responses.append(resp)
        return {"LoadBalancerTargetHealth": responses}

    def _retrieve_load_balancer_listeners(
        self, client: ElasticLoadBalancingv2Client, account_id, region, lbs
    ):
        results = []
        for lb in lbs:
            try:
                # Try a get instead, these don't have to exist

                arn = lb.get("LoadBalancerArn")
                if not arn:
                    print(f"No target group found for {lb['LoadBalancerArn']}")
                    continue
                response = client.describe_listeners(LoadBalancerArn=arn)
                response["LoadBalancerArn"] = lb["LoadBalancerArn"]
                del response["ResponseMetadata"]
            except (ClientError, KeyError) as e:
                self.fatal = True
                logger.error(e)
                return
            results.append(response)
        return {"LoadBalancerListeners": results}

    def _retrieve_load_balancer_attributes(
        self, client: ElasticLoadBalancingv2Client, account_id, region, lbs
    ):
        results = []
        for lb in lbs:
            try:
                arn = lb["LoadBalancerArn"]
                response = client.describe_load_balancer_attributes(LoadBalancerArn=arn)
                response["LoadBalancerArn"] = lb["LoadBalancerArn"]
                del response["ResponseMetadata"]
            except (ClientError, KeyError) as e:
                self.fatal = True
                logger.error(e)
                return
            results.append(response)
        return {"LoadBalancerAttributes": results}

    def get_configs(self, path):
        for account in self.accounts:
            if self.role:
                session = self._assume_role(account["Id"], self.role)
            else:
                session = self._session
            for region in self.regions:
                for _, resource in self.resources.items():
                    client = session.client(resource["client"], region_name=region)
                    self._retrieve_configs(
                        client, resource["function"], account["Id"], path=path
                    )
                self._retrieve_elastic_search_domains(
                    session.client("es", region_name=region), account["Id"], path=path
                )
        if self.fatal:
            raise AWSClientError("Fatal error retreiving AWS configs.")
        return self.fatal



from http import HTTPStatus
from typing import Any, Optional, Union, cast

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.base_error_response import BaseErrorResponse
from ...models.challenge_response import ChallengeResponse
from ...models.modify_allow_inbound_invitations_request import (
    ModifyAllowInboundInvitationsRequest,
)
from ...models.modify_allow_outbound_invitations_request import (
    ModifyAllowOutboundInvitationsRequest,
)
from ...models.modify_default_login_methods_request import (
    ModifyDefaultLoginMethodsRequest,
)
from ...models.validation_error_response import ValidationErrorResponse
from ...types import Response


def _get_kwargs(
    organization_name: str,
    *,
    body: Union[
        "ModifyAllowInboundInvitationsRequest",
        "ModifyAllowOutboundInvitationsRequest",
        "ModifyDefaultLoginMethodsRequest",
    ],
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": "post",
        "url": "/{organization_name}/api/v1/security-settings".format(
            organization_name=organization_name,
        ),
    }

    _body: dict[str, Any]
    if isinstance(body, ModifyAllowInboundInvitationsRequest):
        _body = body.to_dict()
    elif isinstance(body, ModifyAllowOutboundInvitationsRequest):
        _body = body.to_dict()
    else:
        _body = body.to_dict()

    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    if response.status_code == 204:
        response_204 = cast(Any, None)
        return response_204
    if response.status_code == 422:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == 400:
        response_400 = BaseErrorResponse.from_dict(response.json())

        return response_400
    if response.status_code == 401:
        response_401 = ChallengeResponse.from_dict(response.json())

        return response_401
    if response.status_code == 404:
        response_404 = BaseErrorResponse.from_dict(response.json())

        return response_404
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: Union[
        "ModifyAllowInboundInvitationsRequest",
        "ModifyAllowOutboundInvitationsRequest",
        "ModifyDefaultLoginMethodsRequest",
    ],
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Modify security policy

     Modify organization-level security settings. Possible setttings include whether external users can
    be invited into this organization, whether certain login methods are permitted by default, and
    whether users managed by this organization are permitted to participate in outside organizations.

    Args:
        organization_name (str):
        body (Union['ModifyAllowInboundInvitationsRequest',
            'ModifyAllowOutboundInvitationsRequest', 'ModifyDefaultLoginMethodsRequest']):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: Union[
        "ModifyAllowInboundInvitationsRequest",
        "ModifyAllowOutboundInvitationsRequest",
        "ModifyDefaultLoginMethodsRequest",
    ],
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Modify security policy

     Modify organization-level security settings. Possible setttings include whether external users can
    be invited into this organization, whether certain login methods are permitted by default, and
    whether users managed by this organization are permitted to participate in outside organizations.

    Args:
        organization_name (str):
        body (Union['ModifyAllowInboundInvitationsRequest',
            'ModifyAllowOutboundInvitationsRequest', 'ModifyDefaultLoginMethodsRequest']):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
    """

    return sync_detailed(
        organization_name=organization_name,
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: Union[
        "ModifyAllowInboundInvitationsRequest",
        "ModifyAllowOutboundInvitationsRequest",
        "ModifyDefaultLoginMethodsRequest",
    ],
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Modify security policy

     Modify organization-level security settings. Possible setttings include whether external users can
    be invited into this organization, whether certain login methods are permitted by default, and
    whether users managed by this organization are permitted to participate in outside organizations.

    Args:
        organization_name (str):
        body (Union['ModifyAllowInboundInvitationsRequest',
            'ModifyAllowOutboundInvitationsRequest', 'ModifyDefaultLoginMethodsRequest']):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: Union[
        "ModifyAllowInboundInvitationsRequest",
        "ModifyAllowOutboundInvitationsRequest",
        "ModifyDefaultLoginMethodsRequest",
    ],
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Modify security policy

     Modify organization-level security settings. Possible setttings include whether external users can
    be invited into this organization, whether certain login methods are permitted by default, and
    whether users managed by this organization are permitted to participate in outside organizations.

    Args:
        organization_name (str):
        body (Union['ModifyAllowInboundInvitationsRequest',
            'ModifyAllowOutboundInvitationsRequest', 'ModifyDefaultLoginMethodsRequest']):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            organization_name=organization_name,
            client=client,
            body=body,
        )
    ).parsed



import typing

from invariant_client.base_command.base_command import BaseCommand

if typing.TYPE_CHECKING:
    import argparse


class FetchCommand(BaseCommand):
    use_argument_debug = True

    config_path: str | None = None

    @classmethod
    def parse_args(cls, subparsers: 'argparse._SubParsersAction[argparse.ArgumentParser]') -> None:
        command_fetch = subparsers.add_parser(
            'fetch',
            help='Fetch network configs from remote sources to disk.'
        )
        cls._add_common_parser_arguments(command_fetch)

        command_fetch.add_argument(
            'fetch_config',
            nargs='?',
            metavar='config_file',
            default='./invariant_config.yaml',
            help='Fetch configuration file. Configure sources like LibreNMS, AWS, and indicate where and how to find credentials. Default is invariant_config.yaml.'
        )

        command_fetch.add_argument(
            '--output',
            dest='output_path',
            metavar='OUTPUT_PATH',
            help='Download and anonymize configs to a local dir.',
            required=True,
        )

    def set_config(self, args: 'argparse.Namespace', env: dict[str, str]) -> None:
        super().set_config(args, env)
        self.config_path = args.fetch_config
        self.output_path = args.output_path

    def execute(self):
        super().execute()
        self.sdk.fetch(self.config_path, output_path=self.output_path)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.editable_document import EditableDocument
    from ..models.editable_document_version import EditableDocumentVersion


T = TypeVar("T", bound="EditableDocumentResult")


@_attrs_define
class EditableDocumentResult:
    """Result of an editable document read operation, containing the current version and metadata.

    Attributes:
        document (EditableDocument):
        target_version (Union['EditableDocumentVersion', None, Unset]):
        target_contents (Union[None, Unset, str]):
    """

    document: "EditableDocument"
    target_version: Union["EditableDocumentVersion", None, Unset] = UNSET
    target_contents: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.editable_document_version import EditableDocumentVersion

        document = self.document.to_dict()

        target_version: Union[None, Unset, dict[str, Any]]
        if isinstance(self.target_version, Unset):
            target_version = UNSET
        elif isinstance(self.target_version, EditableDocumentVersion):
            target_version = self.target_version.to_dict()
        else:
            target_version = self.target_version

        target_contents: Union[None, Unset, str]
        if isinstance(self.target_contents, Unset):
            target_contents = UNSET
        else:
            target_contents = self.target_contents

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "document": document,
            }
        )
        if target_version is not UNSET:
            field_dict["target_version"] = target_version
        if target_contents is not UNSET:
            field_dict["target_contents"] = target_contents

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.editable_document import EditableDocument
        from ..models.editable_document_version import EditableDocumentVersion

        d = dict(src_dict)
        document = EditableDocument.from_dict(d.pop("document"))

        def _parse_target_version(
            data: object,
        ) -> Union["EditableDocumentVersion", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                target_version_type_0 = EditableDocumentVersion.from_dict(data)

                return target_version_type_0
            except:  # noqa: E722
                pass
            return cast(Union["EditableDocumentVersion", None, Unset], data)

        target_version = _parse_target_version(d.pop("target_version", UNSET))

        def _parse_target_contents(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        target_contents = _parse_target_contents(d.pop("target_contents", UNSET))

        editable_document_result = cls(
            document=document,
            target_version=target_version,
            target_contents=target_contents,
        )

        editable_document_result.additional_properties = d
        return editable_document_result

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.snapshot_metadata import SnapshotMetadata


T = TypeVar("T", bound="UpdatedSnapshotDetail")


@_attrs_define
class UpdatedSnapshotDetail:
    """
    Attributes:
        organization_uuid (UUID):
        network_uuid (UUID):
        metadata (SnapshotMetadata):
        network_name (str):
        uuid (Union[Unset, UUID]):
        created_at (Union[Unset, datetime.datetime]):
        deleted_at (Union[None, Unset, datetime.datetime]):
    """

    organization_uuid: UUID
    network_uuid: UUID
    metadata: "SnapshotMetadata"
    network_name: str
    uuid: Union[Unset, UUID] = UNSET
    created_at: Union[Unset, datetime.datetime] = UNSET
    deleted_at: Union[None, Unset, datetime.datetime] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        organization_uuid = str(self.organization_uuid)

        network_uuid = str(self.network_uuid)

        metadata = self.metadata.to_dict()

        network_name = self.network_name

        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        created_at: Union[Unset, str] = UNSET
        if not isinstance(self.created_at, Unset):
            created_at = self.created_at.isoformat()

        deleted_at: Union[None, Unset, str]
        if isinstance(self.deleted_at, Unset):
            deleted_at = UNSET
        elif isinstance(self.deleted_at, datetime.datetime):
            deleted_at = self.deleted_at.isoformat()
        else:
            deleted_at = self.deleted_at

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "organization_uuid": organization_uuid,
                "network_uuid": network_uuid,
                "metadata": metadata,
                "network_name": network_name,
            }
        )
        if uuid is not UNSET:
            field_dict["uuid"] = uuid
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if deleted_at is not UNSET:
            field_dict["deleted_at"] = deleted_at

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.snapshot_metadata import SnapshotMetadata

        d = dict(src_dict)
        organization_uuid = UUID(d.pop("organization_uuid"))

        network_uuid = UUID(d.pop("network_uuid"))

        metadata = SnapshotMetadata.from_dict(d.pop("metadata"))

        network_name = d.pop("network_name")

        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        _created_at = d.pop("created_at", UNSET)
        created_at: Union[Unset, datetime.datetime]
        if isinstance(_created_at, Unset):
            created_at = UNSET
        else:
            created_at = isoparse(_created_at)

        def _parse_deleted_at(data: object) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                deleted_at_type_0 = isoparse(data)

                return deleted_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        deleted_at = _parse_deleted_at(d.pop("deleted_at", UNSET))

        updated_snapshot_detail = cls(
            organization_uuid=organization_uuid,
            network_uuid=network_uuid,
            metadata=metadata,
            network_name=network_name,
            uuid=uuid,
            created_at=created_at,
            deleted_at=deleted_at,
        )

        updated_snapshot_detail.additional_properties = d
        return updated_snapshot_detail

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from contextlib import contextmanager
import datetime
import io
import json
import logging
import os
import pathlib
import platform
import random
import shutil
import string
import sys
import tempfile
import time
import typing

import backoff
from netconan import anonymize_files
from rich import print_json

from invariant_client import pysdk, zip_util
from invariant_client import display
from invariant_client.aws_pruner.aws_pruner_integration import use_aws_pruner
from invariant_client.base_command.base_command import BaseCommand
from invariant_client.lib.fetcher import FetchManifest
from invariant_client.pysdk import OutputFormat

if typing.TYPE_CHECKING:
    import argparse


logger = logging.getLogger(__name__)


def get_home_directory() -> pathlib.Path:
    """Tries various methods to determine the user's home directory, defaulting to the current working directory."""
    try:
        return pathlib.Path.home()
    except RuntimeError:
        pass

    if platform.system() == 'Windows':
        try:
            return pathlib.Path(os.environ['HOMEDRIVE'], os.environ['HOMEPATH'])
        except KeyError:
            pass

    try:
        return pathlib.Path(os.environ['HOME'])
    except RuntimeError:
        pass

    try:
        return pathlib.Path(os.environ['LAMBDA_TASK_ROOT'])
    except KeyError:
        pass

    return pathlib.Path.cwd()


class RunCommand(BaseCommand):
    use_argument_debug = True
    use_argument_group_format = True
    use_argument_format_condensed = True
    # No TSV format for this command

    @classmethod
    def parse_args(cls, subparsers: 'argparse._SubParsersAction[argparse.ArgumentParser]') -> None:
        command_run = subparsers.add_parser(
            'run',
            description='Analyze the current snapshot.',
            help="Analyze the current snapshot.")

        cls._add_common_parser_arguments(command_run)

        command_run.add_argument(
            '--compare-to',
            dest='compare_to',
            help='Compare this snapshot to another by its git ref. Ref must refer to the primary git repository.',
        )

        command_run.add_argument(
            '--target',
            dest='target',
            help='An Invariant project root directory. Default is current directory.',
        )

        command_run.add_argument(
            '--network',
            dest='network',
            help='The name of the network being evaluated.',
        )

        command_run.add_argument(
            '--role',
            dest='role',
            help='The network role being evaluated, e.g. "live", "intended".',
        )

        # AWS Pruner:
        # The pruner will run if invariant/aws_pruner.yaml exists in the snapshot directory or if --aws-pruner flag is present.
        # The user can disable the pruner with --no-aws-pruner or enabled: False in the config.
        # The user can set --aws-pruner-preview=<dir> to write the pruned snapshot to a directory. It can be combined with --no-aws-pruner or enabled: False .
        # The pruner will write the pruned snapshot to a tempdir and not modify the original.
        command_run.add_argument(
            '--aws-pruner',
            dest='aws_pruner',
            action='store_true',
            help='Activate the AWS pruner.'
        )

        command_run.add_argument(
            '--aws-pruner-dry-run', '--no-aws-pruner',
            dest='no_aws_pruner',
            action='store_true',
            help='AWS pruner runs in dry-run only. Combine with aws-pruner-debug-out to see the would-be pruned snapshot.',
        )

        command_run.add_argument(
            '--aws-pruner-debug-out',
            nargs='?',
            dest='aws_pruner_output_directory',
            const='./aws_pruner_debug',  # Value when --aws-pruner-debug-out is present but has no value
            default=None,  # Value when --aws-pruner-debug-out is NOT present
            help='Specify a directory to write the pruned snapshot for preview purposes. Default is ./aws_pruner_debug/ . Performs a dry-run if --no-aws-pruner is set or the pruner is disabled in aws_pruner.yaml .'
        )

        command_run.add_argument(
            '--no-upload-limit',
            dest='no_upload_limit',
            action='store_true',
            help='Disable the 40MB upload limit for snapshots.',
        )

        command_run.add_argument(
            '--no-wait',
            dest='no_wait',
            action='store_true',
            help='Start the Invariant analysis and exit.',
        )

    def set_config(self, args: 'argparse.Namespace', env: dict[str, str]) -> None:
        super().set_config(args, env)
        self.target = getattr(args, 'target', None) or '.'
        self.network = getattr(args, 'network') or ''
        self.no_upload_limit = getattr(args, 'no_upload_limit', False)
        self.aws_pruner_output_directory = getattr(args, 'aws_pruner_output_directory')
        self.aws_pruner = getattr(args, 'aws_pruner')
        self.no_aws_pruner = getattr(args, 'no_aws_pruner')
        self.no_wait = getattr(args, 'no_wait')

    def execute(self):
        super().execute()

        bytes = None
        if pathlib.Path(self.target).is_file():
            with open(self.target, "rb") as f:
                bytes = io.BytesIO(f.read())
        elif pathlib.Path(self.target).is_dir():
            if not pathlib.Path(self.target, 'configs').is_dir() and not pathlib.Path(self.target, 'aws_configs').is_dir():
                print(f"Invalid directory. Expected subdirectories 'configs' or 'aws_configs' to be present. See https://docs.invariant.tech/Reference/Snapshots for instructions.", file=sys.stderr)
                exit(1)

            home_dir = get_home_directory()
            if pathlib.Path(self.target).absolute() == pathlib.Path(home_dir).absolute():
                print("Upload aborted. Cowardly refusing to upload your home directory.")
                exit(1)

            BYTES_LIMIT = 40000000
            if self.no_upload_limit:
                BYTES_LIMIT = 0

            bytes = io.BytesIO()
            zip_created = False
            if (
                pathlib.Path(self.target, 'aws_configs').is_dir() and \
                (
                    pathlib.Path(self.target, 'invariant', 'aws_pruner.yaml').exists() or \
                    self.aws_pruner or \
                    self.aws_pruner_output_directory
                )
            ):
                with tempfile.TemporaryDirectory() as tempdir:
                    workdir = pathlib.Path(tempdir, pathlib.Path(self.target).absolute().name)
                    shutil.copytree(self.target, workdir)

                    print("Pruner starting...")

                    pruner_debug_target = pathlib.Path(self.target, self.aws_pruner_output_directory) if self.aws_pruner_output_directory else None
                    apply_pruner = use_aws_pruner(workdir, self.no_aws_pruner, pruner_debug_target)
                    if apply_pruner:
                        # Create a new zipfile from the pruned snapshot in the tempdir, discarding the original
                        with anonymized(workdir) as safe_sourcedir:
                            bytes = io.BytesIO()
                            zip_util.zip_dir(safe_sourcedir, bytes, BYTES_LIMIT)
                            zip_created = True
                    else:
                        print(f"Pruner changes discarded (dry run).")

            if not zip_created:
                with anonymized(self.target) as safe_sourcedir:
                    zip_util.zip_dir(safe_sourcedir, bytes, BYTES_LIMIT)  # Write a zip file into 'bytes'
                    zip_created = True
        else:
            print("Unacceptable target", file=sys.stderr)
            print(str(self.target), file=sys.stderr)
            exit(1)

        # Not implemented
        compare_to = None
        role = None

        try:
            exec_uuid = upload_snapshot(self.sdk, bytes, compare_to, self.network, role, self.format)
        except KeyboardInterrupt as e:
            print("Exiting...", file=sys.stderr)
            exit(1)

        if self.format == OutputFormat.TABULATE:
            print("Analysis complete.")
        response = self.sdk.report_detail(exec_uuid)
        # pprint.pprint(asdict(response), width=200)
        if self.format == OutputFormat.JSON or self.format == OutputFormat.FAST_JSON:
            response_json = json.dumps(response.to_dict(), sort_keys=True)
            try:
                print_json(response_json)
            except:
                print(response_json)
        elif self.format == OutputFormat.CONDENSED:
            if response.status['state'] != 'COMPLETE':
                if response.summary['errors'] > 0:
                    errors_locator = response.report.reports.errors
                    errors_response = self.sdk.snapshot_file(errors_locator)
                    display.snapshot_errors(errors_response, self.format)
            display.snapshot_condensed_status(response)
        else:
            display.snapshot_status(response)
            if response.status['state'] == 'COMPLETE':
                display.snapshot_halted(response)
                print('')
                summary = self.sdk.report_detail_text(str(exec_uuid), json_mode=False)
                if summary.text:
                    print(summary.text)
                else:
                    display.snapshot_summary_table(response, self.format)

                print(f"\nRun 'invariant show <file>' to examine any file.")

                if response.summary['errors'] > 0:
                    print(f"\n{response.summary['errors']} {'error' if response.summary['errors'] == 1 else 'errors'} found.")
                    errors_locator = response.report.reports.errors
                    errors_response = self.sdk.snapshot_file(errors_locator)
                    display.snapshot_errors(errors_response, self.format)

            else:
                if response.summary['errors'] > 0:
                    errors_locator = response.report.reports.errors
                    errors_response = self.sdk.snapshot_file(errors_locator)
                    display.snapshot_errors(errors_response, self.format)


@contextmanager
def anonymized(input_path: str | pathlib.Path):
    use_anonymizer = not pathlib.Path(input_path, 'invariant/.fetch.manifest.json').exists()
    if not use_anonymizer:
        yield input_path
        return

    with tempfile.TemporaryDirectory() as temp_path:
        unsafe_path = pathlib.Path(temp_path, 'unsafe')
        safe_path = pathlib.Path(temp_path, 'safe')
        for dir_name in ['aws_configs', 'batfish', 'configs', 'def', 'hosts', 'invariant']:
            input_dir = pathlib.Path(input_path, dir_name)
            unsafe_dir = pathlib.Path(unsafe_path, dir_name)
            if input_dir.exists():
                shutil.copytree(input_dir, unsafe_dir)

        # https://github.com/intentionet/netconan/blob/master/netconan/anonymize_files.py#L36
        salt = "".join(
            random.choice(string.ascii_letters + string.digits) for _ in range(16)
        )
        anonymize_files.anonymize_files(
            unsafe_path,
            safe_path,
            True,  # anonymize passwords
            False,  # anonymize IPs
            salt=salt,
        )
        manifest = FetchManifest().model_dump_json(indent=2)
        pathlib.Path(safe_path, "invariant/").mkdir(parents=True, exist_ok=True)
        pathlib.Path(safe_path, "invariant/.fetch.manifest.json").write_text(manifest)
        yield safe_path


DEFAULT_RETRY_SECONDS = 3


class UploadTerminationError(Exception):
    """An exception that is raised when a snapshot upload is terminated."""

    def __init__(self, *args, retry_after: int):
        super().__init__(self, *args)
        self.retry_after = retry_after


@backoff.on_exception(
        backoff.runtime,
        UploadTerminationError,
        value=lambda e: e.retry_after + random.uniform(0, e.retry_after),
        jitter=None,
        logger=None,
        on_backoff=lambda _: logger.warning('Upload was remotely terminated, retrying...'),
        max_tries=3)
def upload_snapshot(sdk: pysdk.Invariant, bytes: io.BytesIO, compare_to: str, network: str, role: str, format: OutputFormat, no_wait: bool = False) -> str:
    if format == OutputFormat.TABULATE:
        print("Uploading snapshot...")
    exec_uuid = sdk.upload_snapshot(
        source=bytes,
        network=network,
        role=role,
        compare_to=compare_to)
    exec_uuid = exec_uuid.exec_uuid

    if format == OutputFormat.TABULATE:
        print(f"Processing... ({exec_uuid})")
    elif format == OutputFormat.CONDENSED:
        print(f"snapshot: {exec_uuid}")

    if no_wait:
        if format == OutputFormat.TABULATE:
            print("Analysis started in --no_wait mode")
        elif format == OutputFormat.CONDENSED:
            print(f"outcome: started")
        return exec_uuid

    end_time = datetime.datetime.now() + datetime.timedelta(weeks=1)
    while datetime.datetime.now() < end_time:
        response = sdk.upload_is_running(exec_uuid)
        if response.terminated:
            raise UploadTerminationError(f"Upload was remotely terminated, try again later", retry_after=response.retry_after_seconds or DEFAULT_RETRY_SECONDS)
        if not response.is_running:
            break

        # TODO send some RetryAfter header to control this
        # TODO separately, exponential back-off on error
        time.sleep(4)
    if not response:
        print("Timed out.", file=sys.stderr)
        exit(1)
    return exec_uuid


from enum import Enum


class RuleType(str, Enum):
    EGRESS_CRITICAL_FLOW = "egress-critical-flow"
    EGRESS_DENY = "egress-deny"
    EGRESS_DENY_OTHERS = "egress-deny-others"
    INGRESS_CRITICAL_FLOW = "ingress-critical-flow"
    INGRESS_DENY = "ingress-deny"
    INGRESS_DENY_OTHERS = "ingress-deny-others"

    def __str__(self) -> str:
        return str(self.value)



import argparse
import base64
import io
import json
import logging
import pathlib
import re
import sys
import typing
import uuid

from attrs import asdict
import yaml
from dateutil import parser, tz
from rich import print_json
from rich.console import Console
from rich.syntax import Syntax

from invariant_client import display, pysdk, zip_util
from invariant_client.base_command.base_command import BaseCommand
from invariant_client.bindings.invariant_instance_client.models import ExecResponse
from invariant_client.pysdk import OutputFormat
from invariant_client.eval_command.pick_rule import pick_rule
import time


logger = logging.getLogger(__name__)


# Helper type for flattened rules
FlattenedRule = typing.Tuple[str, int, dict] # (policy_name, rule_index, rule_dict)


class EvalCommand(BaseCommand):
    needs_authn = True
    use_argument_debug = True
    use_argument_group_format = True
    use_argument_format_tsv = True
    use_argument_format_condensed = True

    snapshot_uuid: uuid.UUID | None = None
    target: pathlib.Path = pathlib.Path('.')
    rule_file_path: pathlib.Path | None = None
    rule_selector: str | None = None

    @classmethod
    def parse_args(cls, subparsers: 'argparse._SubParsersAction[argparse.ArgumentParser]') -> None:
        command_try_rule = subparsers.add_parser(
            'eval',
            description='Test a single rule against an existing snapshot.',
            help="Test a single access policy rule.")

        cls._add_common_parser_arguments(command_try_rule)

        snapshot_group = command_try_rule.add_mutually_exclusive_group(required=False)
        snapshot_group.add_argument(
            '--snapshot',
            dest='snapshot_uuid',
            help='Specify the snapshot context: a snapshot UUID or a network name (uses latest).',
        )
        snapshot_group.add_argument(
            '--network',
            dest='network_name',
            help='Use the latest snapshot uploaded in the current CLI session.',
        )

        command_try_rule.add_argument(
            '--target',
            dest='target',
            type=pathlib.Path,
            default='.',
            help='An Invariant project root directory containing defs/ and invariant/locations/. Default is current directory.',
        )

        command_try_rule.add_argument(
            '--rule-file',
            dest='rule_file_path',
            type=pathlib.Path,
            required=True,
            help='Path to the YAML file containing the rule to try.',
        )

        command_try_rule.add_argument(
            '--rule',
            dest='rule_selector',
            help='Specify the target rule if --rule-file contains more than one. Format: <policy_name>[<rule_index>] or <global_rule_index>.',
        )

    def set_config(self, args: 'argparse.Namespace', env: dict[str, str]) -> None:
        super().set_config(args, env)
        self.network_name = None
        self.snapshot_uuid = None
        if args.snapshot_uuid:
            self.snapshot_uuid = uuid.UUID(args.snapshot_uuid)
        elif env.get('INVARIANT_SNAPSHOT', None) is not None:
            self.snapshot_uuid = uuid.UUID(env.get('INVARIANT_SNAPSHOT'))
        elif args.network_name:
            self.network_name = args.network_name

        self.target = pathlib.Path(args.target) if args.target is not None else None
        self.rule_file_path = pathlib.Path(args.rule_file_path) if args.rule_file_path is not None else None
        self.rule_selector = args.rule_selector

        if not self.rule_file_path or not self.rule_file_path.is_file():
             raise FileNotFoundError(f"Rule file not found: {self.rule_file_path}")

        if not self.target or not self.target.is_dir():
             raise FileNotFoundError(f"Target not found: {self.target}")


    def _resolve_snapshot_context(self) -> uuid.UUID:
        """Resolves user input into a specific snapshot UUID."""
        if self.snapshot_uuid is None and self.network_name is None:
            # Look up the most recent snapshot for this session
            snapshots_response = self.sdk.list_snapshots(filter_session=True, limit=1)
            if len(snapshots_response) == 0:
                print("No snapshots found for the current session. Please use --network or --snapshot .", file=sys.stderr)
                exit(1)
            snapshot_model = snapshots_response[0].snapshot
            print(f"Using snapshot {snapshot_model.uuid} (uploaded from this session)", file=sys.stderr)
            return snapshot_model.uuid
        elif self.snapshot_uuid is not None:
            # Direct snapshot UUID provided
            return self.snapshot_uuid
        else:
            # Look up the latest snapshot for the specified network name
            snapshots_response = self.sdk.list_snapshots(filter_net=self.network_name, limit=1)
            if len(snapshots_response) == 0:
                print(f"No snapshots found for network {self.network_name}.", file=sys.stderr)
                exit(1)
            snapshot_model = snapshots_response[0].snapshot
            print(f"Using snapshot {snapshot_model.uuid} (latest for network {self.network_name})", file=sys.stderr)
            return snapshot_model.uuid

    def _zip_and_encode(self, directory_path: pathlib.Path) -> str:
        """Zips directory contents in memory and returns base64 encoded string."""
        if not directory_path.is_dir():
            logger.warning(f"Directory not found for zipping: {directory_path}. Skipping.")
            return "" # Return empty string for non-existent directories
        buffer = io.BytesIO()
        try:
            # Assuming zip_util handles empty directories gracefully
            zip_util.zip_dir(str(directory_path), buffer)
            buffer.seek(0)
            zip_content = buffer.read()
            return base64.b64encode(zip_content).decode('utf-8')
        except Exception as e:
            logger.error(f"Failed to zip directory '{directory_path}'", exc_info=e)
            raise RuntimeError(f"Could not process directory: {directory_path}") from e

    def _display_results(
        self,
        response: ExecResponse,
        tested_rule_str: str
        ):
        """Formats and prints the ExecResponse."""

        # Handle Condensed output
        if self.format == OutputFormat.CONDENSED:
            try:
                # A bit naive, assumes violations exist if keys are present and non-empty
                violations_found = False
                for key in ["policy_violations", "critical_flows_violations"]:
                    if key in response.results:
                        # The result is a JSON string representing a list of dicts
                        result_list = json.loads(response.results[key])
                        if isinstance(result_list, list) and len(result_list) > 0:
                             violations_found = True
                             break
                if violations_found:
                    print("FAIL")
                else:
                    # TODO: How does the API indicate an execution error vs. just no violations?
                    # Assuming for now that if no violations, it passed.
                    print("PASS")
            except Exception as e:
                logger.error("Error parsing condensed results", exc_info=e)
                print("ERROR")
            return

        # Handle JSON output first
        # if self.format in (OutputFormat.JSON, OutputFormat.FAST_JSON):
        else:
            output_data = asdict(response)
            try:
                reports = output_data['results'] = output_data['results']['additional_properties']
                parsed_reports = {}
                for key in reports.keys():
                    parsed_reports[key] = json.loads(reports[key])
                output_data['results'] = parsed_reports
            except:
                pass

            try:
                output_data['result_files'] = output_data['result_files']['additional_properties']
            except:
                pass

            if self.format == OutputFormat.FAST_JSON:
                print(json.dumps(output_data, default=str))
            else:
                print_json(json.dumps(output_data, default=str))
            return


    def execute(self) -> None:
        super().execute()
        try:
            resolved_snapshot_uuid = self._resolve_snapshot_context()

            rule_str = pick_rule(self.rule_file_path, self.rule_selector)
            locs_b64 = self._zip_and_encode(self.target.resolve() / 'invariant' / 'locations')
            defs_b64 = self._zip_and_encode(self.target.resolve() / 'def')

            print(f"Trying rule '{self.rule_selector}' from {self.rule_file_path.name} ...", file=sys.stderr)
            # measure the time of try_rule
            start_time = time.time()
            response = self.sdk.try_rule(
                snapshot_uuid=str(resolved_snapshot_uuid),
                rule=rule_str,
                locations=locs_b64,
                defs=defs_b64
            )
            elapsed_time = time.time() - start_time
            print(f"Elapsed time: {elapsed_time:.2f} seconds", file=sys.stderr)
            self._display_results(response, rule_str)

        except (ValueError, FileNotFoundError, yaml.YAMLError, pysdk.RemoteError, pysdk.AuthorizationException) as e:
            print(f"Error: {e}", file=sys.stderr)
            if self.debug:
                logger.exception("Error during try-rule execution")
            sys.exit(1)
        except Exception as e:
            print(f"An unexpected error occurred: {e}", file=sys.stderr)
            if self.debug:
                logger.exception("Unexpected error during try-rule execution")
            sys.exit(1)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.stripe_line_items_response_item_data import (
        StripeLineItemsResponseItemData,
    )
    from ..models.subscription_line_item import SubscriptionLineItem


T = TypeVar("T", bound="StripeLineItemsResponse")


@_attrs_define
class StripeLineItemsResponse:
    """
    Attributes:
        line_items (list['SubscriptionLineItem']):
        item_data (StripeLineItemsResponseItemData):
    """

    line_items: list["SubscriptionLineItem"]
    item_data: "StripeLineItemsResponseItemData"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        line_items = []
        for line_items_item_data in self.line_items:
            line_items_item = line_items_item_data.to_dict()
            line_items.append(line_items_item)

        item_data = self.item_data.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "line_items": line_items,
                "item_data": item_data,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.stripe_line_items_response_item_data import (
            StripeLineItemsResponseItemData,
        )
        from ..models.subscription_line_item import SubscriptionLineItem

        d = dict(src_dict)
        line_items = []
        _line_items = d.pop("line_items")
        for line_items_item_data in _line_items:
            line_items_item = SubscriptionLineItem.from_dict(line_items_item_data)

            line_items.append(line_items_item)

        item_data = StripeLineItemsResponseItemData.from_dict(d.pop("item_data"))

        stripe_line_items_response = cls(
            line_items=line_items,
            item_data=item_data,
        )

        stripe_line_items_response.additional_properties = d
        return stripe_line_items_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import yaml
from abc import ABC, abstractmethod
from urllib.parse import urlparse, parse_qs
import boto3
import os
import json


class Loader(ABC):
    @abstractmethod
    def load(self, uri: str) -> dict:
        """Load and return parsed YAML/JSON content from the given URI."""
        pass

class EnvLoader(Loader):
    def load(self, uri: str) -> dict:
        """
        Load environment variables from a URI like:
        env://VAR_NAME
        env://VAR_NAME?default=default_value
        """
        parsed = urlparse(uri)
        var_name = parsed.netloc
        query = parse_qs(parsed.query)
        default_value = query.get("default", [None])[0]

        value = os.getenv(var_name, default_value)
        if value is None:
            raise ValueError(f"Environment variable '{var_name}' not found and no default provided")

        return yaml.safe_load(value)

class FileLoader(Loader):
    def load(self, uri: str) -> dict:
        path = uri[len("file://") :] if uri.startswith("file://") else uri
        with open(path, "r") as f:
            return yaml.safe_load(f)


class S3Loader(Loader):
    def load(self, uri: str) -> dict:
        parsed = urlparse(uri)
        bucket = parsed.netloc
        key = parsed.path.lstrip("/")
        s3 = boto3.client("s3")
        obj = s3.get_object(Bucket=bucket, Key=key)
        content = obj["Body"].read().decode()
        return yaml.safe_load(content)


class AWSSecretLoader(Loader):
    def load(self, uri: str) -> dict:
        """
        Supports URI like:
        secret+aws://my-secret-name
        secret+aws://my-secret-name?region=us-east-1&version_id=...&version_stage=...&key=myfield
        """
        parsed = urlparse(uri)
        secret_id = parsed.netloc + parsed.path  # combine in case path present
        # secret_id = secret_id.lstrip("/")
        query = parse_qs(parsed.query)
        print(f"query: {query}")

        region = query.get("region", [None])[0]
        version_id = query.get("version_id", [None])[0]
        version_stage = query.get("version_stage", [None])[0]
        key = query.get("key", [None])[0]

        # Create client with region if specified, else default
        if region:
            client = boto3.client("secretsmanager", region_name=region)
        else:
            client = boto3.client("secretsmanager")

        # Build get_secret_value params
        params = {"SecretId": secret_id}
        if version_id:
            params["VersionId"] = version_id
        if version_stage:
            params["VersionStage"] = version_stage
        print(f"params: {params}")
        response = client.get_secret_value(**params)

        secret_string = response.get("SecretString")
        if secret_string is None:
            raise ValueError(f"Secret {secret_id} does not have a SecretString")

        # Parse secret string as JSON or YAML
        try:
            secret_data = json.loads(secret_string)
        except json.JSONDecodeError:
            # Fallback to YAML parsing
            secret_data = yaml.safe_load(secret_string)

        if key:
            # Extract nested key
            for part in key.split("."):
                secret_data = secret_data.get(part)
                if secret_data is None:
                    raise KeyError(f"Key '{key}' not found in secret '{secret_id}'")
            return secret_data

        return secret_data


# Registry mapping scheme -> Loader instance
LOADERS = {
    "file": FileLoader(),
    "s3": S3Loader(),
    "secret+aws": AWSSecretLoader(),
    "env": EnvLoader(),
}

def load(uri: str) -> dict:
    # Defaults to returning string if no scheme is provided
    if "://" not in uri:
        return uri
    scheme = uri.split("://")[0]
    loader = LOADERS.get(scheme)
    if loader is None:
        raise ValueError(f"No loader registered for scheme '{scheme}'")
    return loader.load(uri)



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.policy_file_result import PolicyFileResult
    from ..models.rule_summary_stats import RuleSummaryStats


T = TypeVar("T", bound="RuleSummaryGroupUploaded")


@_attrs_define
class RuleSummaryGroupUploaded:
    """
    Attributes:
        stats (RuleSummaryStats):
        name (str):
        policies (list['PolicyFileResult']):
        type_ (Union[Literal['uploaded'], Unset]):  Default: 'uploaded'.
    """

    stats: "RuleSummaryStats"
    name: str
    policies: list["PolicyFileResult"]
    type_: Union[Literal["uploaded"], Unset] = "uploaded"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        stats = self.stats.to_dict()

        name = self.name

        policies = []
        for policies_item_data in self.policies:
            policies_item = policies_item_data.to_dict()
            policies.append(policies_item)

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "stats": stats,
                "name": name,
                "policies": policies,
            }
        )
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.policy_file_result import PolicyFileResult
        from ..models.rule_summary_stats import RuleSummaryStats

        d = dict(src_dict)
        stats = RuleSummaryStats.from_dict(d.pop("stats"))

        name = d.pop("name")

        policies = []
        _policies = d.pop("policies")
        for policies_item_data in _policies:
            policies_item = PolicyFileResult.from_dict(policies_item_data)

            policies.append(policies_item)

        type_ = cast(Union[Literal["uploaded"], Unset], d.pop("type", UNSET))
        if type_ != "uploaded" and not isinstance(type_, Unset):
            raise ValueError(f"type must match const 'uploaded', got '{type_}'")

        rule_summary_group_uploaded = cls(
            stats=stats,
            name=name,
            policies=policies,
            type_=type_,
        )

        rule_summary_group_uploaded.additional_properties = d
        return rule_summary_group_uploaded

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union, cast

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.base_error_response import BaseErrorResponse
from ...models.body_import_editable_document_organization_name_api_v1_edoc_import_post import (
    BodyImportEditableDocumentOrganizationNameApiV1EdocImportPost,
)
from ...models.challenge_response import ChallengeResponse
from ...models.document_type import DocumentType
from ...models.resource_type import ResourceType
from ...models.validation_error_response import ValidationErrorResponse
from ...types import UNSET, Response, Unset


def _get_kwargs(
    organization_name: str,
    *,
    body: BodyImportEditableDocumentOrganizationNameApiV1EdocImportPost,
    resource_set_name: str,
    document_type: DocumentType,
    resource_type: ResourceType,
    file_path: Union[Unset, str] = "placeholder",
    description: Union[None, Unset, str] = UNSET,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    params: dict[str, Any] = {}

    params["resource_set_name"] = resource_set_name

    json_document_type = document_type.value
    params["document_type"] = json_document_type

    json_resource_type = resource_type.value
    params["resource_type"] = json_resource_type

    params["file_path"] = file_path

    json_description: Union[None, Unset, str]
    if isinstance(description, Unset):
        json_description = UNSET
    else:
        json_description = description
    params["description"] = json_description

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": "post",
        "url": "/{organization_name}/api/v1/edoc/import".format(
            organization_name=organization_name,
        ),
        "params": params,
    }

    _body = body.to_multipart()

    _kwargs["files"] = _body

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    if response.status_code == 204:
        response_204 = cast(Any, None)
        return response_204
    if response.status_code == 422:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == 401:
        response_401 = ChallengeResponse.from_dict(response.json())

        return response_401
    if response.status_code == 404:
        response_404 = BaseErrorResponse.from_dict(response.json())

        return response_404
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: BodyImportEditableDocumentOrganizationNameApiV1EdocImportPost,
    resource_set_name: str,
    document_type: DocumentType,
    resource_type: ResourceType,
    file_path: Union[Unset, str] = "placeholder",
    description: Union[None, Unset, str] = UNSET,
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Import a file based policy for use in the web.

    Args:
        organization_name (str):
        resource_set_name (str):
        document_type (DocumentType):
        resource_type (ResourceType):
        file_path (Union[Unset, str]):  Default: 'placeholder'.
        description (Union[None, Unset, str]):
        body (BodyImportEditableDocumentOrganizationNameApiV1EdocImportPost):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        body=body,
        resource_set_name=resource_set_name,
        document_type=document_type,
        resource_type=resource_type,
        file_path=file_path,
        description=description,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: BodyImportEditableDocumentOrganizationNameApiV1EdocImportPost,
    resource_set_name: str,
    document_type: DocumentType,
    resource_type: ResourceType,
    file_path: Union[Unset, str] = "placeholder",
    description: Union[None, Unset, str] = UNSET,
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Import a file based policy for use in the web.

    Args:
        organization_name (str):
        resource_set_name (str):
        document_type (DocumentType):
        resource_type (ResourceType):
        file_path (Union[Unset, str]):  Default: 'placeholder'.
        description (Union[None, Unset, str]):
        body (BodyImportEditableDocumentOrganizationNameApiV1EdocImportPost):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
    """

    return sync_detailed(
        organization_name=organization_name,
        client=client,
        body=body,
        resource_set_name=resource_set_name,
        document_type=document_type,
        resource_type=resource_type,
        file_path=file_path,
        description=description,
    ).parsed


async def asyncio_detailed(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: BodyImportEditableDocumentOrganizationNameApiV1EdocImportPost,
    resource_set_name: str,
    document_type: DocumentType,
    resource_type: ResourceType,
    file_path: Union[Unset, str] = "placeholder",
    description: Union[None, Unset, str] = UNSET,
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Import a file based policy for use in the web.

    Args:
        organization_name (str):
        resource_set_name (str):
        document_type (DocumentType):
        resource_type (ResourceType):
        file_path (Union[Unset, str]):  Default: 'placeholder'.
        description (Union[None, Unset, str]):
        body (BodyImportEditableDocumentOrganizationNameApiV1EdocImportPost):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        body=body,
        resource_set_name=resource_set_name,
        document_type=document_type,
        resource_type=resource_type,
        file_path=file_path,
        description=description,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: BodyImportEditableDocumentOrganizationNameApiV1EdocImportPost,
    resource_set_name: str,
    document_type: DocumentType,
    resource_type: ResourceType,
    file_path: Union[Unset, str] = "placeholder",
    description: Union[None, Unset, str] = UNSET,
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Import a file based policy for use in the web.

    Args:
        organization_name (str):
        resource_set_name (str):
        document_type (DocumentType):
        resource_type (ResourceType):
        file_path (Union[Unset, str]):  Default: 'placeholder'.
        description (Union[None, Unset, str]):
        body (BodyImportEditableDocumentOrganizationNameApiV1EdocImportPost):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            organization_name=organization_name,
            client=client,
            body=body,
            resource_set_name=resource_set_name,
            document_type=document_type,
            resource_type=resource_type,
            file_path=file_path,
            description=description,
        )
    ).parsed



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.exec_response_result_files import ExecResponseResultFiles
    from ..models.exec_response_results import ExecResponseResults


T = TypeVar("T", bound="ExecResponse")


@_attrs_define
class ExecResponse:
    """
    Attributes:
        exec_uuid (UUID):
        snapshot_uuid (UUID):
        results (ExecResponseResults):
        result_files (ExecResponseResultFiles):
        elapsed (float):
    """

    exec_uuid: UUID
    snapshot_uuid: UUID
    results: "ExecResponseResults"
    result_files: "ExecResponseResultFiles"
    elapsed: float
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        exec_uuid = str(self.exec_uuid)

        snapshot_uuid = str(self.snapshot_uuid)

        results = self.results.to_dict()

        result_files = self.result_files.to_dict()

        elapsed = self.elapsed

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "exec_uuid": exec_uuid,
                "snapshot_uuid": snapshot_uuid,
                "results": results,
                "result_files": result_files,
                "elapsed": elapsed,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.exec_response_result_files import ExecResponseResultFiles
        from ..models.exec_response_results import ExecResponseResults

        d = dict(src_dict)
        exec_uuid = UUID(d.pop("exec_uuid"))

        snapshot_uuid = UUID(d.pop("snapshot_uuid"))

        results = ExecResponseResults.from_dict(d.pop("results"))

        result_files = ExecResponseResultFiles.from_dict(d.pop("result_files"))

        elapsed = d.pop("elapsed")

        exec_response = cls(
            exec_uuid=exec_uuid,
            snapshot_uuid=snapshot_uuid,
            results=results,
            result_files=result_files,
            elapsed=elapsed,
        )

        exec_response.additional_properties = d
        return exec_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.report_task import ReportTask


T = TypeVar("T", bound="ListReportTasksResponse")


@_attrs_define
class ListReportTasksResponse:
    """
    Attributes:
        in_progress (list['ReportTask']):
    """

    in_progress: list["ReportTask"]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        in_progress = []
        for in_progress_item_data in self.in_progress:
            in_progress_item = in_progress_item_data.to_dict()
            in_progress.append(in_progress_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "in_progress": in_progress,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.report_task import ReportTask

        d = dict(src_dict)
        in_progress = []
        _in_progress = d.pop("in_progress")
        for in_progress_item_data in _in_progress:
            in_progress_item = ReportTask.from_dict(in_progress_item_data)

            in_progress.append(in_progress_item)

        list_report_tasks_response = cls(
            in_progress=in_progress,
        )

        list_report_tasks_response.additional_properties = d
        return list_report_tasks_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypeVar,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.integration_data_slack_app_installation_data import (
        IntegrationDataSlackAppInstallationData,
    )


T = TypeVar("T", bound="IntegrationDataSlackAppInstallation")


@_attrs_define
class IntegrationDataSlackAppInstallation:
    """
    Attributes:
        type_ (Literal['slack_app_installation']):
        slack_app_install (IntegrationDataSlackAppInstallationData):
    """

    type_: Literal["slack_app_installation"]
    slack_app_install: "IntegrationDataSlackAppInstallationData"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        type_ = self.type_

        slack_app_install = self.slack_app_install.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type_,
                "slack_app_install": slack_app_install,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.integration_data_slack_app_installation_data import (
            IntegrationDataSlackAppInstallationData,
        )

        d = dict(src_dict)
        type_ = cast(Literal["slack_app_installation"], d.pop("type"))
        if type_ != "slack_app_installation":
            raise ValueError(
                f"type must match const 'slack_app_installation', got '{type_}'"
            )

        slack_app_install = IntegrationDataSlackAppInstallationData.from_dict(
            d.pop("slack_app_install")
        )

        integration_data_slack_app_installation = cls(
            type_=type_,
            slack_app_install=slack_app_install,
        )

        integration_data_slack_app_installation.additional_properties = d
        return integration_data_slack_app_installation

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="BaseErrorResponse")


@_attrs_define
class BaseErrorResponse:
    """Based on RFC7807 - see BaseResponse.

    Attributes:
        status (int):
        type_ (Union[None, str]):
        title (str):
        detail (str):
        instance (Union[None, Unset, str]):
    """

    status: int
    type_: Union[None, str]
    title: str
    detail: str
    instance: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        status = self.status

        type_: Union[None, str]
        type_ = self.type_

        title = self.title

        detail = self.detail

        instance: Union[None, Unset, str]
        if isinstance(self.instance, Unset):
            instance = UNSET
        else:
            instance = self.instance

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "status": status,
                "type": type_,
                "title": title,
                "detail": detail,
            }
        )
        if instance is not UNSET:
            field_dict["instance"] = instance

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        status = d.pop("status")

        def _parse_type_(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        type_ = _parse_type_(d.pop("type"))

        title = d.pop("title")

        detail = d.pop("detail")

        def _parse_instance(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        instance = _parse_instance(d.pop("instance", UNSET))

        base_error_response = cls(
            status=status,
            type_=type_,
            title=title,
            detail=detail,
            instance=instance,
        )

        base_error_response.additional_properties = d
        return base_error_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from urllib.parse import urljoin
import httpx
import sys


class LibreNMSClient:
    """Grabs hosts from the LibreNMS device inventory and then logs into
    each device and gathers configs.
    """

    def __init__(self, base_url: str, api_key: str):
        self.base_url = urljoin(base_url, "/api/v0/")
        headers = {"Authorization": f"Bearer {api_key}"}
        self.client = httpx.Client(headers=headers)
        try:
            self.test_connection()
        except ConnectionError as e:
            raise ConnectionError(e)

    def _request(self, method: str, endpoint: str, **kwargs: dict):
        """Base request method."""
        try:
            url = urljoin(self.base_url, endpoint)
            response = self.client.request(method, url, **kwargs)
            response.raise_for_status()
            return response.json()
        except httpx.ConnectTimeout:
            raise ConnectionError(
                f"Unable to connect to {self.base_url} due to connection timeout."
            )
        except httpx.HTTPStatusError as e:
            raise ConnectionError(f"Error connecting to endpoint: {e}")

    def _get(self, endpoint: str):
        return self._request("GET", endpoint, params={"format": "json"})

    def test_connection(self):
        return self._get("system")

    def get_device(self, node_id: str):
        return self._get(f"devices/{node_id}")

    def get_devices_by_group(self, group: str):
        try:
            result = self._get(f"devicegroups/{group}")
        except ConnectionError as e:
            if "404 Not Found" in str(e):
                print(f"Error, device group {group} not found.")
            else:
                print(e)
            sys.exit(1)
        return result



from enum import Enum


class DocumentType(str, Enum):
    AWS_DESCRIBE = "AWS_DESCRIBE"
    BATFISH_JSON = "BATFISH_JSON"
    INVARIANT_YAML_SPEC = "INVARIANT_YAML_SPEC"
    NET_CONFIG = "NET_CONFIG"
    RESOURCE_SET = "RESOURCE_SET"

    def __str__(self) -> str:
        return str(self.value)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.updated_snapshot_detail import UpdatedSnapshotDetail


T = TypeVar("T", bound="UpdateSnapshotsResponse")


@_attrs_define
class UpdateSnapshotsResponse:
    """
    Attributes:
        snapshots (list['UpdatedSnapshotDetail']):
    """

    snapshots: list["UpdatedSnapshotDetail"]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        snapshots = []
        for snapshots_item_data in self.snapshots:
            snapshots_item = snapshots_item_data.to_dict()
            snapshots.append(snapshots_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "snapshots": snapshots,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.updated_snapshot_detail import UpdatedSnapshotDetail

        d = dict(src_dict)
        snapshots = []
        _snapshots = d.pop("snapshots")
        for snapshots_item_data in _snapshots:
            snapshots_item = UpdatedSnapshotDetail.from_dict(snapshots_item_data)

            snapshots.append(snapshots_item)

        update_snapshots_response = cls(
            snapshots=snapshots,
        )

        update_snapshots_response.additional_properties = d
        return update_snapshots_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.invariant_price import InvariantPrice

T = TypeVar("T", bound="InvariantPlanResponse")


@_attrs_define
class InvariantPlanResponse:
    """
    Attributes:
        value (str):
        prices (list[InvariantPrice]):
        description (str):
        has_capacity (bool):
        hidden (bool):
        permit_enterprise (bool):
        permit_non_enterprise (bool):
    """

    value: str
    prices: list[InvariantPrice]
    description: str
    has_capacity: bool
    hidden: bool
    permit_enterprise: bool
    permit_non_enterprise: bool
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        value = self.value

        prices = []
        for prices_item_data in self.prices:
            prices_item = prices_item_data.value
            prices.append(prices_item)

        description = self.description

        has_capacity = self.has_capacity

        hidden = self.hidden

        permit_enterprise = self.permit_enterprise

        permit_non_enterprise = self.permit_non_enterprise

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "value": value,
                "prices": prices,
                "description": description,
                "has_capacity": has_capacity,
                "hidden": hidden,
                "permit_enterprise": permit_enterprise,
                "permit_non_enterprise": permit_non_enterprise,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        value = d.pop("value")

        prices = []
        _prices = d.pop("prices")
        for prices_item_data in _prices:
            prices_item = InvariantPrice(prices_item_data)

            prices.append(prices_item)

        description = d.pop("description")

        has_capacity = d.pop("has_capacity")

        hidden = d.pop("hidden")

        permit_enterprise = d.pop("permit_enterprise")

        permit_non_enterprise = d.pop("permit_non_enterprise")

        invariant_plan_response = cls(
            value=value,
            prices=prices,
            description=description,
            has_capacity=has_capacity,
            hidden=hidden,
            permit_enterprise=permit_enterprise,
            permit_non_enterprise=permit_non_enterprise,
        )

        invariant_plan_response.additional_properties = d
        return invariant_plan_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class InvariantPrice(str, Enum):
    CAPACITY_ENTERPRISE_TIERED = "capacity_enterprise_tiered"
    CAPACITY_STANDARD_TIERED = "capacity_standard_tiered"
    PRO_PLAN_MONTHLY = "pro_plan_monthly"
    USAGE_ENTERPRISE_METERED = "usage_enterprise_metered"
    USAGE_STANDARD_METERED = "usage_standard_metered"

    def __str__(self) -> str:
        return str(self.value)



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.basic_auth_login_method import BasicAuthLoginMethod
    from ..models.public import Public


T = TypeVar("T", bound="AuthnChallenge")


@_attrs_define
class AuthnChallenge:
    """The user must provide a primary authentication credential.

    Attributes:
        type_ (Literal['authn']):
        allowed_methods (list[Union['BasicAuthLoginMethod', 'Public']]):
    """

    type_: Literal["authn"]
    allowed_methods: list[Union["BasicAuthLoginMethod", "Public"]]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.basic_auth_login_method import BasicAuthLoginMethod

        type_ = self.type_

        allowed_methods = []
        for allowed_methods_item_data in self.allowed_methods:
            allowed_methods_item: dict[str, Any]
            if isinstance(allowed_methods_item_data, BasicAuthLoginMethod):
                allowed_methods_item = allowed_methods_item_data.to_dict()
            else:
                allowed_methods_item = allowed_methods_item_data.to_dict()

            allowed_methods.append(allowed_methods_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type_,
                "allowed_methods": allowed_methods,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.basic_auth_login_method import BasicAuthLoginMethod
        from ..models.public import Public

        d = dict(src_dict)
        type_ = cast(Literal["authn"], d.pop("type"))
        if type_ != "authn":
            raise ValueError(f"type must match const 'authn', got '{type_}'")

        allowed_methods = []
        _allowed_methods = d.pop("allowed_methods")
        for allowed_methods_item_data in _allowed_methods:

            def _parse_allowed_methods_item(
                data: object,
            ) -> Union["BasicAuthLoginMethod", "Public"]:
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    allowed_methods_item_type_0 = BasicAuthLoginMethod.from_dict(data)

                    return allowed_methods_item_type_0
                except:  # noqa: E722
                    pass
                if not isinstance(data, dict):
                    raise TypeError()
                allowed_methods_item_type_1 = Public.from_dict(data)

                return allowed_methods_item_type_1

            allowed_methods_item = _parse_allowed_methods_item(
                allowed_methods_item_data
            )

            allowed_methods.append(allowed_methods_item)

        authn_challenge = cls(
            type_=type_,
            allowed_methods=allowed_methods,
        )

        authn_challenge.additional_properties = d
        return authn_challenge

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import sys
import uuid

from rich import print_json
from invariant_client import display
from invariant_client.base_command.base_command import BaseCommand

import typing

from invariant_client.bindings.invariant_instance_client.models.file_index import FileIndex
from invariant_client.bindings.invariant_instance_client.models.snapshot_report_data import SnapshotReportData
from invariant_client.pysdk import OutputFormat
if typing.TYPE_CHECKING:
    import argparse


class ShowCommand(BaseCommand):
    use_argument_debug = True
    use_argument_group_format = True
    use_argument_format_tsv = True
    use_argument_format_condensed = True

    @classmethod
    def parse_args(cls, subparsers: 'argparse._SubParsersAction[argparse.ArgumentParser]') -> None:
        command_show = subparsers.add_parser(
            'show',
            description='Access network snapshot analysis results.',
            help="Access network snapshot analysis results.")

        cls._add_common_parser_arguments(command_show)

        command_show.add_argument(
            'file_name',
            nargs="?",
            help='The snapshot file to examine.'
        )

        command_show.add_argument(
            '--snapshot',
            dest='snapshot_name',
            help='The snapshot to examine. If unset, environment variable INVARIANT_SNAPSHOT is used.'
        )

    def set_config(self, args: 'argparse.Namespace', env: dict[str, str]) -> None:
        super().set_config(args, env)
        env_snapshot = env.get('INVARIANT_SNAPSHOT', None)
        self.snapshot_name = args.snapshot_name
        if not self.snapshot_name:
            self.snapshot_name = env_snapshot
        self.file_name = args.file_name

    def execute(self):
        super().execute()
        snapshot_name = self.snapshot_name
        if not snapshot_name:
            # NOTE: API Token users should explicitly set --snapshot or INVARIANT_SNAPSHOT
            last_snapshot = self.sdk.list_reports(filter_session=True, limit=1)
            if not last_snapshot or len(last_snapshot.reports) == 0:
                raise ValueError(f"Use --snapshot <name> argument or INVARIANT_SNAPSHOT environment variable to select a snapshot.")
            snapshot_name = last_snapshot.reports[0].uuid

        if isinstance(snapshot_name, uuid.UUID):
            exec_uuid = snapshot_name
        else:
            try:
                exec_uuid = uuid.UUID(snapshot_name, version=4)
            except ValueError as e:
                raise ValueError(f"Expected {snapshot_name} to be a UUID like f5b4e387-e336-499e-b3a0-d6186c590572.") from e

        if self.file_name is not None:
            # Access a specific file
            try:
                file = uuid.UUID(self.file_name, version=4)
            except ValueError:
                # OK if the file is the file key (e.g. errors)
                file = self.file_name
            if not isinstance(file, uuid.UUID):
                # Resolve non-UUID file to UUID
                response = self.sdk.report_detail(exec_uuid)
                reports = response.report.reports
                try:
                    if isinstance(reports, SnapshotReportData):
                        try:
                            file_locator = reports.files[file]
                        except KeyError as e:
                            if file == 'files':
                                raise
                            # Metafile
                            file_locator: FileIndex | uuid.UUID = getattr(reports, file)
                    else:
                        file_locator: uuid.UUID = getattr(reports, file)
                except (KeyError, AttributeError) as e:
                    raise ValueError(f"Report {file} not found for snapshot {exec_uuid}.") from e

            if self.format == OutputFormat.JSON or self.format == OutputFormat.FAST_JSON:
                file_data = self.sdk.snapshot_file(file_locator)
                if self.format == OutputFormat.JSON:
                    print_json(file_data.to_json(orient='records'))
                elif self.format == OutputFormat.FAST_JSON:
                    print(file_data.to_json(orient='records'))

            elif self.format == OutputFormat.TSV:
                file_data = self.sdk.snapshot_file(file_locator)
                display.print_frame(file_data, self.format)
            else:
                file_data = self.sdk.snapshot_file(file_locator)
                display.print_frame(file_data, self.format)
                # print("Set --traces to display all example traces")
                print("Set --json to get JSON")
                print("See 'show --help' for more options")

        else:
            # Display the process summary for the snapshot
            if self.format == OutputFormat.TABULATE:
                print(f"Snapshot {exec_uuid}")
            elif self.format == OutputFormat.CONDENSED:
                print(f"snapshot: {exec_uuid}")
            response = self.sdk.report_detail(str(exec_uuid))
            if self.format == OutputFormat.TABULATE:
                display.snapshot_status(response)
            if self.format == OutputFormat.CONDENSED:
                if response.status['state'] != 'COMPLETE':
                    if response.summary['errors'] > 0:
                        errors_locator = response.report.reports.errors
                        errors_response = self.sdk.snapshot_file(errors_locator)
                        display.snapshot_errors(errors_response, self.format)
                display.snapshot_condensed_status(response)
            elif response.status['state'] == 'COMPLETE':
                if self.format == OutputFormat.TABULATE:
                    display.snapshot_halted(response)
                    print('')
                    summary = self.sdk.report_detail_text(str(exec_uuid), json_mode=False)
                    if summary.text:
                        print(summary.text)
                    else:
                        display.snapshot_summary_table(response, self.format)
                elif self.format == OutputFormat.JSON or self.format == OutputFormat.FAST_JSON:
                    summary = self.sdk.report_detail_text(str(exec_uuid), json_mode=True)
                    if summary.json:
                        if self.format == OutputFormat.JSON:
                            try:
                                print_json(summary.json)
                            except:
                                print(summary.json)
                        elif self.format == OutputFormat.FAST_JSON:
                            print(summary.json)
                    else:
                        display.snapshot_summary_table(response, self.format)
                else:
                    display.snapshot_summary_table(response, self.format)
                if self.format == OutputFormat.TABULATE:
                    print(f"\nRun 'invariant show <file>' to examine any file.")

                    if response.summary['errors'] > 0:
                        print(f"\n{response.summary['errors']} {'error' if response.summary['errors'] == 1 else 'errors'} found.", file=sys.stderr)
                        errors_locator = response.report.reports.errors
                        errors_response = self.sdk.snapshot_file(errors_locator)
                        display.snapshot_errors(errors_response, self.format)
            elif response.summary['errors'] > 0:
                errors_locator = response.report.reports.errors
                errors_response = self.sdk.snapshot_file(errors_locator)
                display.snapshot_errors(errors_response, self.format)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.resource_set import ResourceSet
    from ..models.resource_set_member import ResourceSetMember


T = TypeVar("T", bound="GetResourceSetResponseWithExtras")


@_attrs_define
class GetResourceSetResponseWithExtras:
    """
    Attributes:
        resource_set (ResourceSet):
        current_version (int):
        member_resources (list['ResourceSetMember']):
        attached_networks (list[str]):
    """

    resource_set: "ResourceSet"
    current_version: int
    member_resources: list["ResourceSetMember"]
    attached_networks: list[str]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        resource_set = self.resource_set.to_dict()

        current_version = self.current_version

        member_resources = []
        for member_resources_item_data in self.member_resources:
            member_resources_item = member_resources_item_data.to_dict()
            member_resources.append(member_resources_item)

        attached_networks = self.attached_networks

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "resource_set": resource_set,
                "current_version": current_version,
                "member_resources": member_resources,
                "attached_networks": attached_networks,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.resource_set import ResourceSet
        from ..models.resource_set_member import ResourceSetMember

        d = dict(src_dict)
        resource_set = ResourceSet.from_dict(d.pop("resource_set"))

        current_version = d.pop("current_version")

        member_resources = []
        _member_resources = d.pop("member_resources")
        for member_resources_item_data in _member_resources:
            member_resources_item = ResourceSetMember.from_dict(
                member_resources_item_data
            )

            member_resources.append(member_resources_item)

        attached_networks = cast(list[str], d.pop("attached_networks"))

        get_resource_set_response_with_extras = cls(
            resource_set=resource_set,
            current_version=current_version,
            member_resources=member_resources,
            attached_networks=attached_networks,
        )

        get_resource_set_response_with_extras.additional_properties = d
        return get_resource_set_response_with_extras

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import argparse
import logging
import logging.config
import os
import sys
from typing import Type

from invariant_client.base_command.base_command import BaseCommand
from invariant_client.login_command.login import LoginCommand
from invariant_client.run_command.run import RunCommand
from invariant_client.fetch_command.fetch import FetchCommand
from invariant_client.sync_command.sync import SyncCommand
from invariant_client.show_command.show import ShowCommand
from invariant_client.snapshots_command.snapshots import SnapshotsCommand
from invariant_client.eval_command.eval import EvalCommand
from invariant_client.version_command.version import VersionCommand
from invariant_client.network_command.network import NetworkCommand
from invariant_client.rules_command.rules import RulesCommand
from invariant_client.definition_command.definitions import DefinitionsCommand
from invariant_client.location_command.locations import LocationsCommand


logger = logging.getLogger(__name__)


COMMANDS: dict[str, Type[BaseCommand]] = {
    'login': LoginCommand,
    'eval': EvalCommand,
    'run': RunCommand,
    'sync': SyncCommand,
    'fetch': FetchCommand,
    'show': ShowCommand,
    'snapshots': SnapshotsCommand,
    'networks': NetworkCommand,
    'version': VersionCommand,
    'rules': RulesCommand,
    'definitions': DefinitionsCommand,
    'locations': LocationsCommand,
}
LOGGER_NAME = "invariant"


def parse_args():
    parser = argparse.ArgumentParser(
        prog='invariant',
        description='Invariant analyzes network snapshots',
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparsers = parser.add_subparsers(
        title='available commands',
        description='Run [command] --help for more information.',
        dest='command')

    for command_cls in COMMANDS.values():
        command_cls.parse_args(subparsers)

    parser.add_argument(
        '--version',
        dest='version',
        action='store_true',
        help="Display the client and server version.")

    args = parser.parse_args()

    command = getattr(args, 'command')
    if not command and not args.version:
        parser.print_help()
        exit(0)

    return args

class ModulePathFilter(logging.Filter):
    """
    A logging filter that drops records from a specific module path.
    """
    def __init__(self, block, name=""):
        super().__init__(name)
        self.block = block

    def filter(self, record):
        """
        Blocks a record if its pathname starts with the block and the level is WARNING or below.
        Returns False to block the record, True to allow it.
        """
        # Only filter if the log level is WARNING or below
        return not (
            record.levelno <= logging.WARNING and
            self.block and any(record.pathname.startswith(block_path) for block_path in self.block)
        )

def configure_logging(args):
    debug = getattr(args, 'debug', False)
    verbose = getattr(args, 'verbose', False)
    if debug:
        log_level = logging.DEBUG
    elif verbose:
        log_level = logging.INFO
    else:
        log_level = logging.WARNING
    handlers_config = {}
    formatters_config = {}
    try:
        import rich.console
        import rich.logging
        handlers_config["rich_handler"] = {
            "class": "rich.logging.RichHandler",
            "level": log_level,
            "rich_tracebacks": True,
            "omit_repeated_times": False,
            "tracebacks_show_locals": True,
            "tracebacks_suppress": [],
            "show_time": True,
            "show_path": False,
            "console": rich.console.Console(stderr=True)
        }
    except ImportError:
        formatters_config["custom_console_formatter"] = {
            "format": "%(asctime)s - %(name)s - %(levelname)-8s - %(module)s:%(lineno)d - %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S"
        }
        handlers_config["console_handler"] = {
            "class": "logging.StreamHandler",
            "level": log_level,  # Handler level
            "formatter": "custom_console_formatter", # Use the custom formatter
            "stream": "ext://sys.stderr"
        }

    dict_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "handlers": handlers_config,
        "root": {
            "level": log_level,
            "handlers": list(handlers_config.keys())
        },
        "formatters": formatters_config
    }
    logging.config.dictConfig(dict_config)

    if log_level >= logging.WARNING:
        # Override noisy modules:
        # - netconan: generates warnings when eliminating lines, which is debugging-relevant but not a true warning
        noisy_modules = []
        try:
            import netconan
            # This gives us the directory path of the 'netconan' library
            noisy_modules.append(os.path.dirname(netconan.__file__))

        except (ImportError, AttributeError) as e:
            pass

        if noisy_modules:
            noisy_filter = ModulePathFilter(block=noisy_modules)
            root_logger = logging.getLogger()
            for handler in root_logger.handlers:
                handler.addFilter(noisy_filter)


def EntryPoint():
    args = parse_args()
    configure_logging(args)
    
    try:
        EntryPoint_inner(args)
    except Exception as e:
        if getattr(args, 'debug', False):
            raise e
        print('Error: %s' % e, file=sys.stderr)
        exit(1)


def EntryPoint_inner(args):
    command = getattr(args, 'command')

    if getattr(args, 'version', False):
        command = 'version'

    if command not in COMMANDS:
        print(f"Unknown command {command}", file=sys.stderr)
        exit(1)
    command_inst = COMMANDS[command]()
    env = dict(os.environ)
    command_inst.set_config(args, env)
    command_inst.authenticate()
    command_inst.execute()

    # TODO add test-rule command:
    # Does NOT upload anything (just submits the rule)
    # Waits for the rule to be processed and displays some result info - possibly a table similar to 'run'
    # Question: should the CLI get into rule and network management? E.g. invariant network list, invariant network rules list, etc.

    # TODO now that non-blocking uploads are supported, consider making some changes:
    # 1. Wait for the latest snapshot to finish (same functionality as 'run'), e.g.
    #     print("Processing...")
    #     while datetime.datetime.now() < end_time:
    #         response = sdk.upload_is_running(exec_uuid)



from collections.abc import Mapping
from typing import (
    Any,
    Literal,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="SystemAccessActor")


@_attrs_define
class SystemAccessActor:
    """
    Attributes:
        context (str):
        type_ (Union[Literal['system'], Unset]):  Default: 'system'.
        uuid (Union[Unset, UUID]):  Default: UUID('deadbeef-cafe-431a-aac7-4aecc20c0316').
    """

    context: str
    type_: Union[Literal["system"], Unset] = "system"
    uuid: Union[Unset, UUID] = UUID("deadbeef-cafe-431a-aac7-4aecc20c0316")
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        context = self.context

        type_ = self.type_

        uuid: Union[Unset, str] = UNSET
        if not isinstance(self.uuid, Unset):
            uuid = str(self.uuid)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "context": context,
            }
        )
        if type_ is not UNSET:
            field_dict["type"] = type_
        if uuid is not UNSET:
            field_dict["uuid"] = uuid

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        context = d.pop("context")

        type_ = cast(Union[Literal["system"], Unset], d.pop("type", UNSET))
        if type_ != "system" and not isinstance(type_, Unset):
            raise ValueError(f"type must match const 'system', got '{type_}'")

        _uuid = d.pop("uuid", UNSET)
        uuid: Union[Unset, UUID]
        if isinstance(_uuid, Unset):
            uuid = UNSET
        else:
            uuid = UUID(_uuid)

        system_access_actor = cls(
            context=context,
            type_=type_,
            uuid=uuid,
        )

        system_access_actor.additional_properties = d
        return system_access_actor

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="CreateIntegrationRequestSlackAppInstallationData")


@_attrs_define
class CreateIntegrationRequestSlackAppInstallationData:
    """
    Attributes:
        code (str):
        state (str):
    """

    code: str
    state: str
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        code = self.code

        state = self.state

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "code": code,
                "state": state,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        code = d.pop("code")

        state = d.pop("state")

        create_integration_request_slack_app_installation_data = cls(
            code=code,
            state=state,
        )

        create_integration_request_slack_app_installation_data.additional_properties = d
        return create_integration_request_slack_app_installation_data

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.base_error_response import BaseErrorResponse
from ...models.challenge_response import ChallengeResponse
from ...models.create_managed_user_request import CreateManagedUserRequest
from ...models.create_member_response import CreateMemberResponse
from ...models.create_service_account_request import CreateServiceAccountRequest
from ...models.create_service_account_response import CreateServiceAccountResponse
from ...models.invite_response import InviteResponse
from ...models.invite_user_request import InviteUserRequest
from ...models.validation_error_response import ValidationErrorResponse
from ...types import Response


def _get_kwargs(
    organization_name: str,
    *,
    body: Union[
        "CreateManagedUserRequest", "CreateServiceAccountRequest", "InviteUserRequest"
    ],
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": "post",
        "url": "/{organization_name}/api/v1/members/".format(
            organization_name=organization_name,
        ),
    }

    _body: dict[str, Any]
    if isinstance(body, CreateManagedUserRequest):
        _body = body.to_dict()
    elif isinstance(body, InviteUserRequest):
        _body = body.to_dict()
    else:
        _body = body.to_dict()

    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        Union["CreateMemberResponse", "CreateServiceAccountResponse", "InviteResponse"],
        ValidationErrorResponse,
    ]
]:
    if response.status_code == 200:

        def _parse_response_200(
            data: object,
        ) -> Union[
            "CreateMemberResponse", "CreateServiceAccountResponse", "InviteResponse"
        ]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                response_200_type_0 = CreateMemberResponse.from_dict(data)

                return response_200_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                response_200_type_1 = InviteResponse.from_dict(data)

                return response_200_type_1
            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            response_200_type_2 = CreateServiceAccountResponse.from_dict(data)

            return response_200_type_2

        response_200 = _parse_response_200(response.json())

        return response_200
    if response.status_code == 422:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == 400:
        response_400 = BaseErrorResponse.from_dict(response.json())

        return response_400
    if response.status_code == 401:
        response_401 = ChallengeResponse.from_dict(response.json())

        return response_401
    if response.status_code == 404:
        response_404 = BaseErrorResponse.from_dict(response.json())

        return response_404
    if response.status_code == 409:
        response_409 = BaseErrorResponse.from_dict(response.json())

        return response_409
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        Union["CreateMemberResponse", "CreateServiceAccountResponse", "InviteResponse"],
        ValidationErrorResponse,
    ]
]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: Union[
        "CreateManagedUserRequest", "CreateServiceAccountRequest", "InviteUserRequest"
    ],
) -> Response[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        Union["CreateMemberResponse", "CreateServiceAccountResponse", "InviteResponse"],
        ValidationErrorResponse,
    ]
]:
    """Create or invite user

     Create or invite a user in the organization. Creating a user creates a 'managed user', a whole user
    account managed by this organization. An invited user is not managed by this organization. You can
    disallow outside user invitations in the security settings.

    Args:
        organization_name (str):
        body (Union['CreateManagedUserRequest', 'CreateServiceAccountRequest',
            'InviteUserRequest']):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[BaseErrorResponse, ChallengeResponse, Union['CreateMemberResponse', 'CreateServiceAccountResponse', 'InviteResponse'], ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: Union[
        "CreateManagedUserRequest", "CreateServiceAccountRequest", "InviteUserRequest"
    ],
) -> Optional[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        Union["CreateMemberResponse", "CreateServiceAccountResponse", "InviteResponse"],
        ValidationErrorResponse,
    ]
]:
    """Create or invite user

     Create or invite a user in the organization. Creating a user creates a 'managed user', a whole user
    account managed by this organization. An invited user is not managed by this organization. You can
    disallow outside user invitations in the security settings.

    Args:
        organization_name (str):
        body (Union['CreateManagedUserRequest', 'CreateServiceAccountRequest',
            'InviteUserRequest']):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[BaseErrorResponse, ChallengeResponse, Union['CreateMemberResponse', 'CreateServiceAccountResponse', 'InviteResponse'], ValidationErrorResponse]
    """

    return sync_detailed(
        organization_name=organization_name,
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: Union[
        "CreateManagedUserRequest", "CreateServiceAccountRequest", "InviteUserRequest"
    ],
) -> Response[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        Union["CreateMemberResponse", "CreateServiceAccountResponse", "InviteResponse"],
        ValidationErrorResponse,
    ]
]:
    """Create or invite user

     Create or invite a user in the organization. Creating a user creates a 'managed user', a whole user
    account managed by this organization. An invited user is not managed by this organization. You can
    disallow outside user invitations in the security settings.

    Args:
        organization_name (str):
        body (Union['CreateManagedUserRequest', 'CreateServiceAccountRequest',
            'InviteUserRequest']):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[BaseErrorResponse, ChallengeResponse, Union['CreateMemberResponse', 'CreateServiceAccountResponse', 'InviteResponse'], ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    body: Union[
        "CreateManagedUserRequest", "CreateServiceAccountRequest", "InviteUserRequest"
    ],
) -> Optional[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        Union["CreateMemberResponse", "CreateServiceAccountResponse", "InviteResponse"],
        ValidationErrorResponse,
    ]
]:
    """Create or invite user

     Create or invite a user in the organization. Creating a user creates a 'managed user', a whole user
    account managed by this organization. An invited user is not managed by this organization. You can
    disallow outside user invitations in the security settings.

    Args:
        organization_name (str):
        body (Union['CreateManagedUserRequest', 'CreateServiceAccountRequest',
            'InviteUserRequest']):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[BaseErrorResponse, ChallengeResponse, Union['CreateMemberResponse', 'CreateServiceAccountResponse', 'InviteResponse'], ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            organization_name=organization_name,
            client=client,
            body=body,
        )
    ).parsed



from collections.abc import Mapping
from typing import Any, TypeVar
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="OrganizationSummary")


@_attrs_define
class OrganizationSummary:
    """
    Attributes:
        uuid (UUID):
        name (str):
        description (str):
        color (str):
        icon (str):
        url (str):
    """

    uuid: UUID
    name: str
    description: str
    color: str
    icon: str
    url: str
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        uuid = str(self.uuid)

        name = self.name

        description = self.description

        color = self.color

        icon = self.icon

        url = self.url

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "uuid": uuid,
                "name": name,
                "description": description,
                "color": color,
                "icon": icon,
                "url": url,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        uuid = UUID(d.pop("uuid"))

        name = d.pop("name")

        description = d.pop("description")

        color = d.pop("color")

        icon = d.pop("icon")

        url = d.pop("url")

        organization_summary = cls(
            uuid=uuid,
            name=name,
            description=description,
            color=color,
            icon=icon,
            url=url,
        )

        organization_summary.additional_properties = d
        return organization_summary

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="UsageStats")


@_attrs_define
class UsageStats:
    """Counts usages.

    Attributes:
        periods (list[int]):
        uploads (list[int]):
        pro (list[int]):
        basic (list[int]):
        basic_eval (list[int]):
    """

    periods: list[int]
    uploads: list[int]
    pro: list[int]
    basic: list[int]
    basic_eval: list[int]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        periods = self.periods

        uploads = self.uploads

        pro = self.pro

        basic = self.basic

        basic_eval = self.basic_eval

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "periods": periods,
                "uploads": uploads,
                "pro": pro,
                "basic": basic,
                "basic_eval": basic_eval,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        periods = cast(list[int], d.pop("periods"))

        uploads = cast(list[int], d.pop("uploads"))

        pro = cast(list[int], d.pop("pro"))

        basic = cast(list[int], d.pop("basic"))

        basic_eval = cast(list[int], d.pop("basic_eval"))

        usage_stats = cls(
            periods=periods,
            uploads=uploads,
            pro=pro,
            basic=basic,
            basic_eval=basic_eval,
        )

        usage_stats.additional_properties = d
        return usage_stats

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.email_subscriber import EmailSubscriber
    from ..models.slack_subscriber import SlackSubscriber


T = TypeVar("T", bound="NotificationGroupMetadata")


@_attrs_define
class NotificationGroupMetadata:
    """
    Attributes:
        name (str):
        comment (Union[None, Unset, str]):
        email_targets (Union[None, Unset, list[str]]):
        subscribers (Union[Unset, list[Union['EmailSubscriber', 'SlackSubscriber']]]):
        network_subscriptions (Union[None, Unset, list[str]]):
    """

    name: str
    comment: Union[None, Unset, str] = UNSET
    email_targets: Union[None, Unset, list[str]] = UNSET
    subscribers: Union[Unset, list[Union["EmailSubscriber", "SlackSubscriber"]]] = UNSET
    network_subscriptions: Union[None, Unset, list[str]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.email_subscriber import EmailSubscriber

        name = self.name

        comment: Union[None, Unset, str]
        if isinstance(self.comment, Unset):
            comment = UNSET
        else:
            comment = self.comment

        email_targets: Union[None, Unset, list[str]]
        if isinstance(self.email_targets, Unset):
            email_targets = UNSET
        elif isinstance(self.email_targets, list):
            email_targets = self.email_targets

        else:
            email_targets = self.email_targets

        subscribers: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.subscribers, Unset):
            subscribers = []
            for subscribers_item_data in self.subscribers:
                subscribers_item: dict[str, Any]
                if isinstance(subscribers_item_data, EmailSubscriber):
                    subscribers_item = subscribers_item_data.to_dict()
                else:
                    subscribers_item = subscribers_item_data.to_dict()

                subscribers.append(subscribers_item)

        network_subscriptions: Union[None, Unset, list[str]]
        if isinstance(self.network_subscriptions, Unset):
            network_subscriptions = UNSET
        elif isinstance(self.network_subscriptions, list):
            network_subscriptions = self.network_subscriptions

        else:
            network_subscriptions = self.network_subscriptions

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
            }
        )
        if comment is not UNSET:
            field_dict["comment"] = comment
        if email_targets is not UNSET:
            field_dict["email_targets"] = email_targets
        if subscribers is not UNSET:
            field_dict["subscribers"] = subscribers
        if network_subscriptions is not UNSET:
            field_dict["network_subscriptions"] = network_subscriptions

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.email_subscriber import EmailSubscriber
        from ..models.slack_subscriber import SlackSubscriber

        d = dict(src_dict)
        name = d.pop("name")

        def _parse_comment(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        comment = _parse_comment(d.pop("comment", UNSET))

        def _parse_email_targets(data: object) -> Union[None, Unset, list[str]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                email_targets_type_0 = cast(list[str], data)

                return email_targets_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[str]], data)

        email_targets = _parse_email_targets(d.pop("email_targets", UNSET))

        subscribers = []
        _subscribers = d.pop("subscribers", UNSET)
        for subscribers_item_data in _subscribers or []:

            def _parse_subscribers_item(
                data: object,
            ) -> Union["EmailSubscriber", "SlackSubscriber"]:
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    subscribers_item_type_0 = EmailSubscriber.from_dict(data)

                    return subscribers_item_type_0
                except:  # noqa: E722
                    pass
                if not isinstance(data, dict):
                    raise TypeError()
                subscribers_item_type_1 = SlackSubscriber.from_dict(data)

                return subscribers_item_type_1

            subscribers_item = _parse_subscribers_item(subscribers_item_data)

            subscribers.append(subscribers_item)

        def _parse_network_subscriptions(data: object) -> Union[None, Unset, list[str]]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                network_subscriptions_type_0 = cast(list[str], data)

                return network_subscriptions_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, list[str]], data)

        network_subscriptions = _parse_network_subscriptions(
            d.pop("network_subscriptions", UNSET)
        )

        notification_group_metadata = cls(
            name=name,
            comment=comment,
            email_targets=email_targets,
            subscribers=subscribers,
            network_subscriptions=network_subscriptions,
        )

        notification_group_metadata.additional_properties = d
        return notification_group_metadata

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="SubscriptionLineItem")


@_attrs_define
class SubscriptionLineItem:
    """
    Attributes:
        id (str):
        subscription_id (str):
        product_id (str):
        quantity (int):
        is_original (bool):
    """

    id: str
    subscription_id: str
    product_id: str
    quantity: int
    is_original: bool
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        subscription_id = self.subscription_id

        product_id = self.product_id

        quantity = self.quantity

        is_original = self.is_original

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "subscription_id": subscription_id,
                "product_id": product_id,
                "quantity": quantity,
                "is_original": is_original,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        id = d.pop("id")

        subscription_id = d.pop("subscription_id")

        product_id = d.pop("product_id")

        quantity = d.pop("quantity")

        is_original = d.pop("is_original")

        subscription_line_item = cls(
            id=id,
            subscription_id=subscription_id,
            product_id=product_id,
            quantity=quantity,
            is_original=is_original,
        )

        subscription_line_item.additional_properties = d
        return subscription_line_item

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
